"""
File containing a dictionary which holds the description for all the grid_options settings. 

Using this, together with the accompanied function, the user can interactively check what the settings result in. 

With this its also possible to automatically generate a pdf file containing all the setting names + descriptions. 
"""

grid_options_descriptions = {}

def grid_options_help(option):
    """
    Function that returns the description of a grid option
    """

    pass


def grid_options_description_checker():
    """
    Function that checks which descriptions are missing
    """

    pass